@@ -6,7 +6,7 @@ import multiprocessing
 import os
 import logging
 from tempfile import TemporaryDirectory
-from typing import Optional, Iterator, Tuple, Callable, Any
+from typing import Dict, Optional, Iterator, Tuple, Callable, Any
 
 import faiss
 import fsspec
@@ -27,9 +27,13 @@ from autofaiss.utils.decorators import Timeit
 logger = logging.getLogger("autofaiss")
 
 
-def _generate_small_index_file_name(batch_id: int, nb_batches: int) -> str:
+def _generate_suffix(batch_id: int, nb_batches: int) -> str:
     suffix_width = int(math.log10(nb_batches)) + 1
-    return "index_" + str(batch_id).zfill(suffix_width)
+    return str(batch_id).zfill(suffix_width)
+
+
+def _generate_small_index_file_name(batch_id: int, nb_batches: int) -> str:
+    return "index_" + _generate_suffix(batch_id, nb_batches)
 
 
 def _save_small_index(index: faiss.Index, batch_id: int, small_indices_folder: str, nb_batches: int) -> None:
@@ -128,21 +132,21 @@ def _batch_loader(batch_size: int, nb_batches: int) -> Iterator[Tuple[int, int,
         yield batch_id, start, end
 
 
-def _get_stage2_folder(tmp_folder: str) -> str:
-    """Get the temporary folder path used for the 2nd stage of merging"""
-    return tmp_folder.rstrip("/") + "/stage-2/"
-
-
 def _merge_index(
     small_indices_folder: str,
     nb_batches: int,
     batch_id: Optional[int] = None,
     start: Optional[int] = None,
     end: Optional[int] = None,
-    max_size_on_disk: str = "100GB",
+    max_size_on_disk: str = "50GB",
     tmp_output_folder: Optional[str] = None,
-) -> faiss.Index:
-    """Merge all the indices in `small_indices_folder` into single one return the merged index."""
+    index_optimizer: Callable = None,
+) -> Tuple[faiss.Index, Dict[str, str]]:
+    """
+    Merge all the indices in `small_indices_folder` into single one.
+    Also run optimization when `index_optimizer` is given.
+    Returns the merged index and the metric
+    """
     fs = _get_file_system(small_indices_folder)
     small_indices_files = sorted(fs.ls(small_indices_folder, detail=False))
     small_indices_files = small_indices_files[start:end]
@@ -172,7 +176,7 @@ def _merge_index(
     first_index_file = small_indices_files[0]
     first_index_size = fs.size(first_index_file)
     max_sizes_in_bytes = cast_memory_to_bytes(max_size_on_disk)
-    nb_files_each_time = math.floor(max_sizes_in_bytes / first_index_size)
+    nb_files_each_time = max(1, int(max_sizes_in_bytes / first_index_size))
     merged_index = None
     n = len(small_indices_files)
     nb_iterations = max(math.ceil(n / nb_files_each_time), 1)
@@ -188,43 +192,59 @@ def _merge_index(
                 pbar.update(1)
 
     if batch_id is not None and tmp_output_folder is not None:
-        _save_small_index(
-            index=merged_index, batch_id=batch_id, small_indices_folder=tmp_output_folder, nb_batches=nb_batches
-        )
-    return merged_index
+        if index_optimizer is not None:
+            metric_infos = index_optimizer(merged_index, index_suffix=_generate_suffix(batch_id, nb_batches))
+        else:
+            metric_infos = None
+            _save_small_index(
+                index=merged_index, batch_id=batch_id, small_indices_folder=tmp_output_folder, nb_batches=nb_batches
+            )
+    else:
+        metric_infos = None
+    return merged_index, metric_infos
 
 
 def _get_file_system(path: str) -> fsspec.AbstractFileSystem:
-    return fsspec.core.url_to_fs(path)[0]
+    return fsspec.core.url_to_fs(path, use_listings_cache=False)[0]
 
 
-def _merge_to_n_indices(spark_session, n: int, src_folder: str):
-    """Merge all the indices from src_folder into n indices, and return the folder for the next stage"""
+def _merge_to_n_indices(spark_session, n: int, src_folder: str, dst_folder: str, index_optimizer: Callable = None):
+    """Merge all the indices from src_folder into n indices, and return the folder for the next stage, as well as the metrics"""
     fs = _get_file_system(src_folder)
     nb_indices_on_src_folder = len(fs.ls(src_folder, detail=False))
-    if nb_indices_on_src_folder <= n:
-        # merging does not happen, return the source folder
-        return src_folder
-    dst_folder = _get_stage2_folder(src_folder)
-    batch_size = math.ceil(nb_indices_on_src_folder / n)
+
+    if nb_indices_on_src_folder <= n and index_optimizer is None:
+        # no need to merge
+        return src_folder, None
+
+    batch_size = max(1, int(nb_indices_on_src_folder / n))
     n = math.ceil(nb_indices_on_src_folder / batch_size)
     merge_batches = _batch_loader(batch_size=batch_size, nb_batches=n)
+
     rdd = spark_session.sparkContext.parallelize(merge_batches, n)
-    rdd.foreach(
-        lambda x: _merge_index(
+
+    def merge(x):
+        _, metrics = _merge_index(
             small_indices_folder=src_folder,
             nb_batches=n,
             batch_id=x[0],
             start=x[1],
             end=x[2],
             tmp_output_folder=dst_folder,
+            index_optimizer=index_optimizer,
         )  # type: ignore
-    )
-    fs = _get_file_system(src_folder)
+        return metrics
+
+    metrics_rdd = rdd.map(merge)
+    metrics = list(metrics_rdd.collect())
+    if index_optimizer is not None:
+        metrics_dict = {metric_info["index_path"]: metric_info for metric_info in metrics}  # type: ignore
+    else:
+        metrics_dict = None  # type: ignore
     for file in fs.ls(src_folder, detail=False):
         if fs.isfile(file):
             fs.rm(file)
-    return dst_folder
+    return dst_folder, metrics_dict
 
 
 def run(
@@ -235,7 +255,8 @@ def run(
     temporary_indices_folder="hdfs://root/tmp/distributed_autofaiss_indices",
     embedding_ids_df_handler: Optional[Callable[[pd.DataFrame, int], Any]] = None,
     nb_indices_to_keep: int = 1,
-) -> Tuple[Optional[faiss.Index], Optional[str]]:
+    index_optimizer: Optional[Callable] = None,
+) -> Tuple[Optional[faiss.Index], Optional[Dict[str, str]]]:
     """
     Create indices by pyspark.
 
@@ -255,12 +276,14 @@ def run(
         The function that handles the embeddings Ids when id_columns is given
     nb_indices_to_keep: int
         Number of indices to keep at most after the merging step
+    index_optimizer: Optional[Callable]
+        The function that optimizes the index
     """
     temporary_indices_folder = make_path_absolute(temporary_indices_folder)
     fs = _get_file_system(temporary_indices_folder)
     if fs.exists(temporary_indices_folder):
         fs.rm(temporary_indices_folder, recursive=True)
-
+    stage1_folder = temporary_indices_folder.rstrip("/") + "/stage-1"
     ss = _get_pyspark_active_session()
     # broadcast the index bytes
     trained_index_bytes = get_bytes_from_index(faiss_index)
@@ -270,7 +293,7 @@ def run(
 
     # maximum between the number of spark workers, 100M embeddings per task and the number of indices to keep
     estimated_nb_batches = max(n_workers, int(embedding_reader.count / (10 ** 7)), nb_indices_to_keep)
-    batch_size = math.ceil(embedding_reader.count / estimated_nb_batches)
+    batch_size = max(1, int(embedding_reader.count / estimated_nb_batches))
     nb_batches = math.ceil(embedding_reader.count / batch_size)
 
     batches = _batch_loader(batch_size=batch_size, nb_batches=nb_batches)
@@ -284,7 +307,7 @@ def run(
                 memory_available_for_adding=memory_available_for_adding,
                 broadcast_trained_index_bytes=broadcast_trained_index_bytes,
                 embedding_reader=embedding_reader,
-                small_indices_folder=temporary_indices_folder,
+                small_indices_folder=stage1_folder,
                 num_cores=num_cores_per_executor,
                 embedding_ids_df_handler=embedding_ids_df_handler,
                 nb_batches=nb_batches,
@@ -292,13 +315,25 @@ def run(
         )
 
     with Timeit("-> Merging indices", indent=2):
-        next_stage_folder = _merge_to_n_indices(spark_session=ss, n=100, src_folder=temporary_indices_folder,)
+        stage2_folder = temporary_indices_folder.rstrip("/") + "/stage-2"
+        next_stage_folder, _ = _merge_to_n_indices(
+            spark_session=ss, n=100, src_folder=stage1_folder, dst_folder=stage2_folder, index_optimizer=None
+        )
         if nb_indices_to_keep == 1:
-            merged_index = _merge_index(small_indices_folder=next_stage_folder, nb_batches=1)
-            fs.rm(temporary_indices_folder, recursive=True)
-            return merged_index, None
+            merged_index, _ = _merge_index(small_indices_folder=next_stage_folder, nb_batches=1)
+            if fs.exists(temporary_indices_folder):
+                fs.rm(temporary_indices_folder, recursive=True)
+            metrics = index_optimizer(merged_index, "")  # type: ignore
+            return merged_index, metrics
         else:
-            next_stage_folder = _merge_to_n_indices(
-                spark_session=ss, n=nb_indices_to_keep, src_folder=next_stage_folder,
+            final_folder = temporary_indices_folder.rstrip("/") + "/final"
+            next_stage_folder, metrics = _merge_to_n_indices(
+                spark_session=ss,
+                n=nb_indices_to_keep,
+                src_folder=next_stage_folder,
+                dst_folder=final_folder,
+                index_optimizer=index_optimizer,
             )
-            return None, next_stage_folder
+            if fs.exists(temporary_indices_folder):
+                fs.rm(temporary_indices_folder, recursive=True)
+            return None, metrics
