@@ -2,10 +2,13 @@
 
 import functools
 import time
+import logging
 from contextlib import ContextDecorator
 from datetime import datetime
 from typing import Optional
 
+logger = logging.getLogger("autofaiss")
+
 
 class Timeit(ContextDecorator):
     """Timing class, used as a context manager"""
@@ -21,7 +24,10 @@ class Timeit(ContextDecorator):
             if self.comment is not None:
                 space = "\t" * self.indent
                 start_date = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
-                print(f"{space}{self.comment} {start_date}", flush=True)
+                logger.info(f"{space}{self.comment} {start_date}")
+                # flush to make sure we display log in stdout before entering in the wrapped function
+                for h in logger.handlers:
+                    h.flush()
             self.start_time = time.perf_counter()
 
     def __exit__(self, *exc):
@@ -29,7 +35,7 @@ class Timeit(ContextDecorator):
             end_time = time.perf_counter()
             run_time = end_time - self.start_time
             space = "\t" * self.indent
-            print(f'{space}>>> Finished "{self.comment}" in {run_time:.4f} secs')
+            logger.info(f'{space}>>> Finished "{self.comment}" in {run_time:.4f} secs')
 
 
 timeit = Timeit()
@@ -44,7 +50,7 @@ def timer(func):
         value = func(*args, **kwargs)
         end_time = time.perf_counter()
         run_time = end_time - start_time
-        print(f"Finished {func.__name__!r} in {run_time:.4f} secs")
+        logger.info(f"Finished {func.__name__!r} in {run_time:.4f} secs")
         return value
 
     return wrapper_timer
